--**************************************************************************
--
--   Copyright (C) 1991-1994 Lotus Development Corporation.
--   All rights reserved.
--
--   1.  Lotus grants you a non-exclusive royalty-free right to use and modify
--       the source code version and to reproduce and distribute the object
--       code version of the Example Programs provided in the Toolkit, provided
--       that you:
--
--       a) distribute the Example Programs only in conjunction with and as a
--          part of your software application product which is designed as an
--          add-in to Lotus 1-2-3, and
--
--       b) do not represent your product to be a product of Lotus Development
--          Corporation.
--
--   2.  The Example Programs are provided as is, without warranty of any kind,
--       either express or implied, including, without limitation, the implied
--       warranties of merchantability of fitness for a particular purpose.
--
--   File Name:  date.pl
--
--**************************************************************************

-- This module implements the Date ADT
-- Date has the following attributes:
--    Day, Month, Year, DateNumber, and Note.
-- Day, Month, and Year describe the day, month, and year
-- a Date object represents.  DateNumber contains a date serial number
-- (numbering begins at 1, which represents January 1, 1900).
-- The Note attribute lets you read and modify notes for a particular date.
-- For instance, a note might contain the string "Don't forget anniversary".
-- The Date ADT also has a twenty-four-hour appointment calendar so you can
-- read and modify appointments for a particular date.


use atdate -- library file that has date procedures

-------------------------------------------------
--         CONSTANTS
-------------------------------------------------

constant NoteNumber = 10   -- default number of notes
constant ApptNumber = 24   -- default number of appointments

--------------------------------------------------
--         TYPES
--------------------------------------------------

-- used in String procedures to specify the date format
public type FormatType is oneof
               (FullYear, FullDate, DayMonth, MonthYear)

-- used in Date definition; specifies
-- an appointment and appointment time
public type Appointment is record
     string Time, Appoint
     end record

-------------------------------------------------
--          ABSTRACT TYPE
-------------------------------------------------

-- defines the underlying structure of Date
abstract type Date [NoteNum, ApptNum] is record
     integer  Day
     integer  Month
     integer  Year
     integer  DateNumber   -- date serial number
     string array [NoteNum] Notes
     Appointment array [ApptNum] Schedule
     end record

------------------------------------------------------
--         PROCEDURE CALLED BY CONSTRUCTORS
------------------------------------------------------

-- InitializeTime initializes the Time fields of
-- the Appointments in the Schedule array so that
-- the schedule has 24 hours.  The array is indexed
-- by the string values assigned to the Time fields,
-- for instance: Date["12am"].
-- Input: a Date to be initialized
-- Returns: the initialized Date
-- Called by the Date constructors

procedure InitializeTime (Date [NoteNum, ApptNum] DateObject)
                          returns (DateObject) is

   DateObject.Schedule[1].Time  = "12am"
   DateObject.Schedule[2].Time  = "1am"
   DateObject.Schedule[3].Time  = "2am"
   DateObject.Schedule[4].Time  = "3am"
   DateObject.Schedule[5].Time  = "4am"
   DateObject.Schedule[6].Time  = "5am"
   DateObject.Schedule[7].Time  = "6am"
   DateObject.Schedule[8].Time  = "7am"
   DateObject.Schedule[9].Time  = "8am"
   DateObject.Schedule[10].Time = "9am"
   DateObject.Schedule[11].Time = "10am"
   DateObject.Schedule[12].Time = "11am"
   DateObject.Schedule[13].Time = "12pm"
   DateObject.Schedule[14].Time = "1pm"
   DateObject.Schedule[15].Time = "2pm"
   DateObject.Schedule[16].Time = "3pm"
   DateObject.Schedule[17].Time = "4pm"
   DateObject.Schedule[18].Time = "5pm"
   DateObject.Schedule[19].Time = "6pm"
   DateObject.Schedule[20].Time  = "7pm"
   DateObject.Schedule[21].Time = "8pm"
   DateObject.Schedule[22].Time = "9pm"
   DateObject.Schedule[23].Time = "10pm"
   DateObject.Schedule[24].Time = "11pm"
   end procedure InitializeTime

-------------------------------------------------------
--              CONSTRUCTOR PROCEDURES
-------------------------------------------------------

-- Date constructs a Date object that represents
-- the specified Day, Month, Year, and number of notes;
-- and has ApptNumber number of appointments.
-- Input: integers for the Day, Month, Year (4-digit year),
--        and number of notes.
-- Returns: a Date object
-- Signals: InvalidDate if the Day, Month, or Year is invalid
--          InvalidNumber if the number of notes is negative
-- Calls the Date library procedure to get the date serial number
--           InitializeTime to initialize the Time fields in Schedule

procedure Date (integer Day, Month, Year, NumNote)
                  returns (Date [NumNote, ApptNumber] DateObject)
                  signals (InvalidDate, InvalidNumber) is

  if NumNote < 1 then     -- only positive numbers
     signal InvalidNumber
  end if
  DateObject.DateNumber = AtDate.Date(Year - 1900, Month, Day)
    except on ERR, NA do  -- signals raised by Date
      signal InvalidDate
      end except
    DateObject.Day = Day
    DateObject.Month = Month
    DateObject.Year = Year
    DateObject.Notes = ""
    for integer i = 1 to ApptNumber loop
      DateObject.Schedule[i].Time = ""
      DateObject.Schedule[i].Appoint = ""
      end loop
    DateObject = InitializeTime(DateObject)
    end procedure Date


-- Constructs a Date object that represents the specified  Day,
-- Month, and Year, and defaults to NoteNumber notes and
-- ApptNumber appointments.
-- Input: integers for the Day, Month, and Year (4-digit year)
-- Returns: a Date object
-- Signals: InvalidDate if the Day, Month, or Year is invalid
-- Calls the Date library procedure to get the date serial number
--         InitializeTime to initialize the Time fields in Schedule

procedure Date (integer Day, Month, Year)
                  returns (Date [NoteNumber, ApptNumber] DateObject)
                  signals (InvalidDate) is

  DateObject.DateNumber = AtDate.Date(Year - 1900, Month, Day)
    except on ERR, NA do   -- signals raised by Date
      signal InvalidDate
      end except
    DateObject.Day = Day
    DateObject.Month = Month
    DateObject.Year = Year
    DateObject.Notes = ""
    for integer i = 1 to ApptNumber loop
       DateObject.Schedule[i].Time = ""
       DateObject.Schedule[i].Appoint = ""
       end loop
    DateObject = InitializeTime(DateObject)
    end procedure Date


-- Constructs a Date object that represents the current date
-- and defaults to NoteNumber notes and ApptNumber number appointments.
-- Input: none
-- Returns: a Date object
-- Signals: InvalidDate if any of the library procedures called
--          raises a signal
-- Calls the library procedures Today, Day, Month, Year, and Date
--         InitializeTime to initialize the Time fields in Schedule

procedure Date  returns (Date [NoteNumber, ApptNumber] DateObject)
                   signals (InvalidDate) is

begin
  DateObject.Day = Day(Today)
  DateObject.Month = Month(Today)
  -- example: 55 + 1900 = 1955
  DateObject.Year = Year(Today) + 1900
  DateObject.DateNumber =
    AtDate.Date(DateObject.Year - 1900, DateObject.Month, DateObject.Day)
end
    except on ERR, NA do   -- signals raised by Day, Month, Year, Date
       signal InvalidDate
       end except
  DateObject.Notes = ""
  for integer i = 1 to ApptNumber loop
     DateObject.Schedule[i].Time = ""
     DateObject.Schedule[i].Appoint = ""
     end loop
  DateObject = InitializeTime(DateObject)
 end procedure Date

------------------------------------------------------
--       ATTRIBUTE PROCEDURES
------------------------------------------------------


-- Read the Day attribute and return the day as an integer.
-- For instance, if the date is 28-apr-89, return 28.
-- Input: a Date object
-- Returns: an integer (the day)
-- Signals: none

procedure Day (Date [NoteNum, ApptNum] ThisDate)
               returns (integer ThisDay) is

  ThisDay = ThisDate.Day
  end procedure Day


-- Read the Day attribute and return the day as the day of the week
-- For instance, if the date is 28-apr-89, return "Friday"
-- Input: a Date object
-- Returns: a string that is the day of the week
-- Signals: none

procedure Day (Date [NoteNum, ApptNum] ThisDate)
               returns (string ThisDay) is

  case ThisDate.DateNumber mod 7
    of 0  : ThisDay = "Saturday"
    of 1  : ThisDay = "Sunday"
    of 2  : ThisDay = "Monday"
    of 3  : ThisDay = "Tuesday"
    of 4  : ThisDay = "Wednesday"
    of 5  : ThisDay = "Thursday"
    of 6  : ThisDay = "Friday"
    end case
  end procedure Day


-- Modify the Day attribute
-- Input: a Date object and an integer (the new day value)
-- Returns: the updated DateObject
-- Signals: InvalidDay if day is invalid
-- Calls the Date library procedure

procedure Day (Date [NoteNum, ApptNum] DateObject, integer NewDay)
               returns (DateObject)
               signals (InvalidDay) is

  DateObject.DateNumber =
    AtDate.Date(DateObject.Year - 1900, DateObject.Month, NewDay)
    except on ERR, NA do  -- signals raised by Date
      signal InvalidDay
      end except
  DateObject.Day = NewDay
  end procedure Day


-- Read the Month attribute and return value as an
-- integer (1 through 12).
-- Input: a Date object
-- Returns: an integer (the month)
-- Signals: none

procedure Month (Date [NoteNum, ApptNum] DateObject)
                 returns (integer ThisMonth) is

  ThisMonth = DateObject.Month
  end procedure Month


-- Read the Month attribute and return the value
-- as a string (January through December)
-- Input: a Date object
-- Returns: a string that is the Month
-- Signals: none

procedure Month (Date [NoteNum, ApptNum] DateObject)
                 returns (string ThisMonth) is

  case DateObject.Month
    of 1  : ThisMonth = "January"
    of 2  : ThisMonth = "February"
    of 3  : ThisMonth = "March"
    of 4  : ThisMonth = "April"
    of 5  : ThisMonth = "May"
    of 6  : ThisMonth = "June"
    of 7  : ThisMonth = "July"
    of 8  : ThisMonth = "August"
    of 9  : ThisMonth = "September"
    of 10 : ThisMonth = "October"
    of 11 : ThisMonth = "November"
    of 12 : ThisMonth = "December"
    end case
  end procedure Month


-- Modify the Month attribute
-- Input: a Date object and an integer (the new month)
-- Returns: the updated DateObject
-- Signals: InvalidMonth if month is invalid
-- Calls the Date library procedure

procedure Month
      (Date [NoteNum, ApptNum] DateObject, integer NewMonth)
       returns (DateObject)
       signals (InvalidMonth) is

  DateObject.DateNumber =
    AtDate.Date(DateObject.Year - 1900, NewMonth, DateObject.Day)
    except on ERR, NA do
      signal InvalidMonth
      end except
  DateObject.Month = NewMonth
  end procedure Month


-- Read the Year attribute
-- Input: a Date object
-- Returns: an integer (the year)
-- Signals: none

procedure Year (Date [NoteNum, ApptNum] DateObject)
                returns (integer ThisYear) is

  ThisYear = DateObject.Year
  end procedure Year


-- Modify the Year attribute
-- Input: a Date object and an integer (the new year)
-- Returns: the updated Date object
-- Signals: InvalidYear if the year is invalid
-- Calls the Date library procedure

procedure Year (Date [NoteNum, ApptNum] DateObject, integer NewYear)
                returns (DateObject)
                signals (InvalidYear) is

    DateObject.DateNumber =
      AtDate.Date(NewYear - 1900, DateObject.Month, DateObject.Day)
      except on ERR, NA  do
        signal InvalidYear
        end except
    DateObject.Year = NewYear
  end procedure Year


-- Read the DateNumber attribute
-- (There is no modify procedure for DateNumber)
-- Input: a Date object
-- Returns: an integer (the date serial number)
-- Signals: none

procedure DateNumber (Date [NoteNum, ApptNum] DateObject)
                      returns (integer ThisNumber) is

  ThisNumber = DateObject.DateNumber
  end procedure DateNumber


-- Read a note in the Note attribute
-- Input: a Date object and an integer (indexing the note to read)
-- Returns: a string (the note)
-- Signals: OutOfBounds if the index value is not in the range
--          of the Note array bounds

procedure Note
      (Date [NoteNum, ApptNum] DateObject, integer WhichNote)
       returns (string ThisNote)
       signals (OutOfBounds) is

  if WhichNote < 1 OR WhichNote > NoteNum then
    signal OutOfBounds
  else
    ThisNote = DateObject.Notes[WhichNote]
  end if
  end procedure Note


-- Modify a note in the Note attribute
-- Input: a Date object, an integer (indexing the note to modify),
--        and a string (the new note)
-- Returns: the updated Date object
-- Signals: OutOfBounds if the index value is not in the range
--          of the Note array bounds

procedure Note (Date [NoteNum, ApptNum] DateObject,
                integer WhichNote, string NewNote)
                returns (DateObject)
                signals (OutOfBounds) is

  if WhichNote < 1 OR WhichNote > NoteNum then
    signal OutOfBounds
  else
    DateObject.Notes[WhichNote] = NewNote
  end if
  end procedure Note

-------------------------------------------------------------
--    OPERATOR PROCEDURES
-------------------------------------------------------------

-- Add a date to an integer to get a date N days from DateObject
-- Input: a DateObject and an integer (the number of days to add)
-- Returns: a new DateObject (that is N days from the original date)
-- Signals: InvalidNumber if a negative integer is passed as input
--          ERR if any of the library procedures called raises
--          a signal
-- Calls the Day, Month, and Year library procedures

procedure Plus (Date [NoteNum, ApptNum] DateObject, integer NDays)
                returns (Date [NoteNum, ApptNum] FutureDate)
                signals (InvalidNumber, ERR) is

  if NDays < 1 then     -- allow only positive numbers
     signal InvalidNumber
  end if
  FutureDate.DateNumber = DateObject.DateNumber + NDays
begin
  FutureDate.Day = Day(FutureDate.DateNumber)
  FutureDate.Month = Month(FutureDate.DateNumber)
  FutureDate.Year = Year(FutureDate.DateNumber) + 1900
end
    except on ERR, NA do   -- signals raised by Day, Month, Year
       signal ERR
       end except
  end procedure Plus

-- Subtract a date from an integer to get the date
-- N days before DateObject.
-- Input: a Date object and an integer (number of days to subtract)
-- Returns: a new date (that is N days before the original date)
-- Signals: InvalidNumber if a negative number is passed as input
--          ERR if any of the library procedures called raises a signal
-- Calls the Day, Month, and Year library procedures

procedure Minus (Date [NoteNum, ApptNum] DateObject, integer NDays)
                 returns (Date [NoteNum, ApptNum] PrevDate)
                 signals (InvalidNumber, ERR) is

  if NDays < 1 then     --  allow only positive numbers
     signal InvalidNumber
  end if
  PrevDate.DateNumber = DateObject.DateNumber - NDays
begin
  PrevDate.Day = Day(PrevDate.DateNumber)
  PrevDate.Month = Month(PrevDate.DateNumber)
  PrevDate.Year = Year(PrevDate.DateNumber) + 1900
end
    except on ERR, NA do   -- signals raised by Day, Month, Year
       signal ERR
       end except
  end procedure Minus


-- Subtract two Dates to find how many days there are between them
-- Input: 2 DateObjects
-- Returns: an integer (the number of days between the 2 dates)
-- Signals: none
-- Calls the Abs built-in procedure

procedure Minus (Date [NoteNum, ApptNum] DateObject1, DateObject2)
                 returns (integer NumDays) is

  -- take the absolute value so NumDays is always positive
  NumDays = Abs(DateObject1.DateNumber - DateObject2.DateNumber)
  end procedure Minus


-- Select an appointment from the Schedule field using a string
-- (e.g. "8am") as an index.
-- The string corresponds to the string in the Time field of
-- an Appointment record.  Example: Date["8am"] refers to
-- the Appoint field value in the record whose Time field = "8am".
-- Input: a DataADT object and a string (indexing the appointment)
-- Returns: a string (the appointment at the specified time)
-- Signals: NoSuchTime if the string index does not correspond to
--          any of the times in Schedule.

procedure Select
      (Date [NoteNum, ApptNum] DateObject, string WhichTime)
       returns (string ThisAppt)
       signals (NoSuchTime) is

   -- Search the Schedule for a time that matches WhichTime.
   -- If found, return the Appoint string
   for integer i = 1 to ApptNumber loop
      if DateObject.Schedule[i].Time = WhichTime then
         ThisAppt = DateObject.Schedule[i].Appoint
         return       -- end procedure if found
         end if
       end loop
   signal NoSuchTime  -- signal if time string not found
       end procedure Select


-- Replace an appointment, in the Schedule field using a string
-- (e.g. "8am") as an index.  The string corresponds to the string
-- in the Time field of an Appointment record.
-- Example: Date["12pm"] = "Lunch with David"
-- Input: a Date object, a string (indexing the appointment),
--        and a string that is the new appointment
-- Returns: the updated Date object
-- Signals: NoSuchTime if the string index does not correspond to
--          any of the times in Schedule

procedure Replace
  (Date [NoteNum, ApptNum] DateObject, string WhichTime, NewAppoint)
   returns (DateObject)
   signals (NoSuchTime) is

   -- Search the Schedule for a time that matches WhichTime.
   -- If found, replace the Appoint string
   for integer i = 1 to ApptNumber loop
      if DateObject.Schedule[i].Time = WhichTime then
         DateObject.Schedule[i].Appoint = NewAppoint
         return       -- end procedure if found
         end if
       end loop
   signal NoSuchTime  -- signal if time string not found
       end procedure Replace

---------------------------------------------------------------
--     PROCEDURES THAT MANIPULATE DATE OBJECTS
---------------------------------------------------------------


-- Return the Date object as a string
-- in the default date format of DD-MMM-YYYY
-- Input: a Date object
-- Returns: a string (the date in string form)
-- Signals: none
-- Calls Month and the String type conversion procedure

procedure String (Date [NoteNum, ApptNum] DateObject)
                  returns (string DateString) is

  string YearString = String(DateObject.Year)
  string MonthString = Month(DateObject)
  string DayString = String(DateObject.Day)
  DateString =
      DayString || "-" || MonthString[1..3] || "-" || YearString
  end procedure String


-- Return the Date object as a string
-- in the specified date format
-- formats:  DD-MMM-YYYY, DD-MMM-YY, DD-MMM, MMM-YY
-- Input: a Date object and the format type (a member of FormatType)
-- Returns: a string (the date in string form)
-- Signals: none
-- Calls Month and the String type conversion procedure

procedure String
      (Date [NoteNum, ApptNum] DateObject, FormatType Format)
       returns (string DateString) is

  -- use strings for the year, month, and day
  string YearString = String(DateObject.Year)
  string MonthString = Month(DateObject)
  string DayString = String(DateObject.Day)

  -- put strings in the specified date format
  case Format
    of FullYear :   DateString = DayString || "-" || MonthString[1..3]
                    || "-" || YearString
    of FullDate :   DateString = DayString || "-" || MonthString[1..3]
                    || "-" || YearString[3..4]
    of DayMonth :   DateString = DayString || "-" || MonthString[1..3]
    of MonthYear:   DateString = MonthString[1..3] || "-" ||
                    YearString[3..4]
    end case
  end procedure String
