--**************************************************************************
--
--   Copyright (C) 1991-1994 Lotus Development Corporation.
--   All rights reserved.
--
--   1.  Lotus grants you a non-exclusive royalty-free right to use and modify
--       the source code version and to reproduce and distribute the object
--       code version of the Example Programs provided in the Toolkit, provided
--       that you:
--
--       a) distribute the Example Programs only in conjunction with and as a
--          part of your software application product which is designed as an
--          add-in to Lotus 1-2-3, and
--
--       b) do not represent your product to be a product of Lotus Development
--          Corporation.
--
--   2.  The Example Programs are provided as is, without warranty of any kind,
--       either express or implied, including, without limitation, the implied
--       warranties of merchantability of fitness for a particular purpose.
--
--   File Name:  nraph.pl
--
--**************************************************************************


--
--  Nraph.PL
--
--  To use: compile with no arguments to demonstrate use of Newton-Raphson
--      method of finding a root using the three methods (C infix, C stack,
--      LPL) of implementation.  Results will be printed in print.log
--
--  Compile with -D DO_TIMING to perform timing tests.  This will result in
--      execution of TIMING_ITERATIONS iterations of root finding for each
--      implementation and print resulting times at the end, along with root
--      results for purposes of comparison.
--


use newtrap     -- contains C infix and stack implementation of N-R
use atdate      -- for timing purposes

---------------------------------------------------------------------------
--
-- Useful constants
--
---------------------------------------------------------------------------
#define TIMING_ITERATIONS   1000    -- number of times to loop for timings

#define COLWIDTH    15              -- output column widths
#define DECIMALS    5               -- # of output decimal places

#define NUMROOTS    6               -- Array sizes, # of roots found in this
                                    -- test program.  If you change this,
                                    -- you must adjust array initialization
                                    -- in main procedure

--
-- Define names for some roots
--
#define SQUAREROOT  2
#define CUBEROOT    3
#define QUADROOT    4
#define QUINTROOT   5
#define SEXTAROOT   6
#define SEPTAROOT   7
#define OCTAROOT    8

---------------------------------------------------------------------------
--
-- Useful Types
--
---------------------------------------------------------------------------
type RootSpec is record     -- holds specification of number to find root
        float   number      -- of and the root to find
        integer root
    end record


public type findroots is                        -- procedure variable for
            procedure (float f1, integer i2)    -- newton raphson procedures
            returns(float Result)


---------------------------------------------------------------------------
--
--  LPL implementation of newton-raphson method for finding a root
--
---------------------------------------------------------------------------
procedure lef_nraph (float Number, integer Index)
    returns ( float Root ) is

    float k, guep
    integer j
   float array [2] guess

   guess[1] = 1.
   guess[2] = 1.

   k = float(index) - 1.
   loop
--      /*substitute x for x sub zero */
      guess[1] = guess[2]

      guep = 1.
      for j = 1 to index - 1 loop
         guep = guep * guess[1] -- /* calculate x sub zero to the index-1 power */
      end loop

--      /* execute Newton-Raphson formula */
      guess[2] = (k * guess[1] + number/guep)/float(index)

   if  Abs(guess[2] - guess[1]) < .0000001 then
        exit
   end if
   end loop


--   /* Test for Epsilon.  A do-while loop is used to make the loop execute
--    * once before evaluation of Epsilon.  Otherwise the loop would not
--    * execute at all, since both the guess variables are initialized at the
--    * same time.  Move one more than the desired number of decimal places
--    * to the left of the decimal point.  Note multiplication by a float value.
--    */

    root = guess[2] * 1000000.


--    /* use integer division and multiplication to reduce units place to zero.
--     * Round up if appropriate.  Tens place in 'root' as it stands here will
--     * become the last decimal place.
--     */
    if (root - (root/10) * 10 >= 5.) then
         root = root + 10.
    end if

--  lop off remaining decimal places since only valid to 5 places
    root = floor(root)

--   /* return root to its correct value, but with last decimal place rounded.
--    * Note division by a float value
--    */
   root = root / 1000000.

end procedure lef_nraph


----------------------------------------------------------------------------
--
--  format strings for nice output
--
----------------------------------------------------------------------------
--
-- Turn floating point # into a string and display N decimal places
--
procedure fformat (float x, integer decplaces)
                    returns (string s) is

    s = string(x)                           -- convert x to a string
    integer i = index(s, ".")               -- find decimal place location
        except on notfound do                   -- if none, return
            return
        end except

    i = i + 1                               -- move to digit following "."
    if (i + decplaces < length(s)) then     -- are we under the length?
        s[i+decplaces..*] = ""              -- if we are, truncate
    end if

end procedure fformat


--
-- Format a float and an integer in a string in columns COLWIDTH wide
--
procedure sformat (float num1, integer num2)
            returns (string fmtstr) is

string  blanks = "                    "

    fmtstr = fformat(num1, DECIMALS)
    fmtstr = fmtstr || blanks[1..(COLWIDTH - length(string(num1)))]
    fmtstr = fmtstr || string(num2) || blanks[1..(COLWIDTH - length(string(num2)))]

end procedure sformat


--
-- Format 3 floats in a string of columns COLWIDTH wide
--
procedure sformat (float num1, num2, num3)
            returns (string fmtstr) is
string  s1 = fformat(num1, DECIMALS)
string  s2 = fformat(num2, DECIMALS)
string  s3 = fformat(num3, DECIMALS)
string  blanks = "                    "

    fmtstr = s1 || blanks[1..(COLWIDTH - length(s1))]
    fmtstr = fmtstr || s2 || blanks[1..(COLWIDTH - length(s2))]
    fmtstr = fmtstr || s3 || blanks[1..(COLWIDTH - length(s3))]

end procedure sformat



----------------------------------------------------------------------------
--
--  Find the Roots
--
----------------------------------------------------------------------------
procedure DoRoots (RootSpec array [rcount] RootArray,
                        Float Array [rcount] ResultArray,
                        FindRoots RootProc)
                returns(ResultArray) is

#if DO_TIMING
    float   starttime, elapsedtime      -- timing variables
    integer x
#endif

#if DO_TIMING
    starttime = now()   -- ok, get the starting time
    for  x = 1 to TIMING_ITERATIONS loop
#endif

        -- loop through the elements of the input array and call RootProc
        -- to fill in the elements of the ResultArray

        for integer j = 1 to RootArray.rcount loop
            ResultArray[j] = RootProc(RootArray[j].number, RootArray[j].root)
        end loop

#if DO_TIMING
    end loop
    elapsedtime = now() - starttime
    print ("Elapsed time = "|| string(minute(elapsedtime))||":"||string(second(elapsedtime)))
#endif


end procedure DoRoots


----------------------------------------------------------------------------
--
--  Main procedure, test preceeding code
--
----------------------------------------------------------------------------
Procedure main is

    RootSpec array [NUMROOTS] RootArray -- array of root specifications

    --
    -- initialize RootArray:
    --
    RootArray[1].number = 4.0
    RootArray[1].root   = QUINTROOT

    RootArray[2].number = 27.0
    RootArray[2].root   = CUBEROOT

    RootArray[3].number = 125.0
    RootArray[3].root   = QUADROOT

    RootArray[4].number = 17.0
    RootArray[4].root   = SQUAREROOT

    RootArray[5].number = 45.273
    RootArray[5].root   = QUADROOT

    RootArray[6].number = 6.76
    RootArray[6].root   = SEXTAROOT

    -- output arrays
    Float Array [NUMROOTS]  infix_C, stackbased_C, infix_LPL


    print ("------------------------------------------------------------------------------")
    print ("Newton-Raphson Method of root approximation:")

-- Using C infix notation
#if DO_TIMING
    print("Infix C based:")
#endif
    infix_C = DoRoots(RootArray, infix_C, findroots(newton_raphson))

-- Using C stack based
#if DO_TIMING
    print("Stack C based:")
#endif
    stackbased_C = DoRoots(RootArray, stackbased_C, findroots(newton_raphson2))


-- Using LPL
#if DO_TIMING
    print("Infix LPL based:")
#endif
    infix_LPL = DoRoots(RootArray, infix_LPL, findroots(lef_nraph))


-- Print the output in a nice format for comparison purposes
    print ("")
    print ("                                              RESULT")
    print ("                              ----------------------------------------")
    print ("Number         Root            C Infix        C Stack          LPL")
    print ("------         ----           -----------    ----------     ----------")
    for integer i = 1 to NUMROOTS loop
        print (sformat(RootArray[i].number, RootArray[i].root) ||
                sformat(infix_C[i], stackbased_C[i], infix_LPL[i]))
    end loop
    print ("------------------------------------------------------------------------------")
    print ("")
    print ("")

end procedure main
